import { Kind, TransformKind } from '../../type/symbols/index.mjs';
import { TypeBoxError } from '../../type/error/index.mjs';
import { KeyOfPropertyKeys } from '../../type/keyof/index.mjs';
import { Index } from '../../type/indexed/index.mjs';
import { Deref } from '../deref/index.mjs';
import { Check } from '../check/index.mjs';
// ------------------------------------------------------------------
// ValueGuard
// ------------------------------------------------------------------
import { IsPlainObject, IsArray, IsValueType } from '../guard/index.mjs';
// ------------------------------------------------------------------
// TypeGuard
// ------------------------------------------------------------------
import { IsTransform, IsSchema } from '../../type/guard/type.mjs';
// ------------------------------------------------------------------
// Errors
// ------------------------------------------------------------------
export class TransformEncodeCheckError extends TypeBoxError {
    schema;
    value;
    error;
    constructor(schema, value, error) {
        super(`Unable to encode due to invalid value`);
        this.schema = schema;
        this.value = value;
        this.error = error;
    }
}
export class TransformEncodeError extends TypeBoxError {
    schema;
    value;
    constructor(schema, value, error) {
        super(`${error instanceof Error ? error.message : 'Unknown error'}`);
        this.schema = schema;
        this.value = value;
    }
}
// ------------------------------------------------------------------
// Encode
// ------------------------------------------------------------------
// prettier-ignore
function Default(schema, value) {
    try {
        return IsTransform(schema) ? schema[TransformKind].Encode(value) : value;
    }
    catch (error) {
        throw new TransformEncodeError(schema, value, error);
    }
}
// prettier-ignore
function FromArray(schema, references, value) {
    const defaulted = Default(schema, value);
    return IsArray(defaulted)
        ? defaulted.map((value) => Visit(schema.items, references, value))
        : defaulted;
}
// prettier-ignore
function FromIntersect(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value) || IsValueType(value))
        return defaulted;
    const knownKeys = KeyOfPropertyKeys(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in defaulted
            ? { ...value, [key]: Visit(Index(schema, [key]), references, value[key]) }
            : value;
    }, defaulted);
    if (!IsTransform(schema.unevaluatedProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const unevaluatedProperties = schema.unevaluatedProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(unevaluatedProperties, value[key]) }
            : value;
    }, knownProperties);
}
// prettier-ignore
function FromNot(schema, references, value) {
    return Default(schema.not, Default(schema, value));
}
// prettier-ignore
function FromObject(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value))
        return defaulted;
    const knownKeys = KeyOfPropertyKeys(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in value
            ? { ...value, [key]: Visit(schema.properties[key], references, value[key]) }
            : value;
    }, defaulted);
    if (!IsSchema(schema.additionalProperties)) {
        return knownProperties;
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
// prettier-ignore
function FromRecord(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value))
        return defaulted;
    const pattern = Object.getOwnPropertyNames(schema.patternProperties)[0];
    const knownKeys = new RegExp(pattern);
    const knownProperties = Object.getOwnPropertyNames(value).reduce((value, key) => {
        return knownKeys.test(key)
            ? { ...value, [key]: Visit(schema.patternProperties[pattern], references, value[key]) }
            : value;
    }, defaulted);
    if (!IsSchema(schema.additionalProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.test(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
// prettier-ignore
function FromRef(schema, references, value) {
    const target = Deref(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
// prettier-ignore
function FromThis(schema, references, value) {
    const target = Deref(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
// prettier-ignore
function FromTuple(schema, references, value) {
    const value1 = Default(schema, value);
    return IsArray(schema.items) ? schema.items.map((schema, index) => Visit(schema, references, value1[index])) : [];
}
// prettier-ignore
function FromUnion(schema, references, value) {
    // test value against union variants
    for (const subschema of schema.anyOf) {
        if (!Check(subschema, references, value))
            continue;
        const value1 = Visit(subschema, references, value);
        return Default(schema, value1);
    }
    // test transformed value against union variants
    for (const subschema of schema.anyOf) {
        const value1 = Visit(subschema, references, value);
        if (!Check(schema, references, value1))
            continue;
        return Default(schema, value1);
    }
    return Default(schema, value);
}
// prettier-ignore
function Visit(schema, references, value) {
    const references_ = typeof schema.$id === 'string' ? [...references, schema] : references;
    const schema_ = schema;
    switch (schema[Kind]) {
        case 'Array':
            return FromArray(schema_, references_, value);
        case 'Intersect':
            return FromIntersect(schema_, references_, value);
        case 'Not':
            return FromNot(schema_, references_, value);
        case 'Object':
            return FromObject(schema_, references_, value);
        case 'Record':
            return FromRecord(schema_, references_, value);
        case 'Ref':
            return FromRef(schema_, references_, value);
        case 'This':
            return FromThis(schema_, references_, value);
        case 'Tuple':
            return FromTuple(schema_, references_, value);
        case 'Union':
            return FromUnion(schema_, references_, value);
        default:
            return Default(schema_, value);
    }
}
/**
 * `[Internal]` Encodes the value and returns the result. This function expects the
 * caller to pass a statically checked value. This function does not check the encoded
 * result, meaning the result should be passed to `Check` before use. Refer to the
 * `Value.Encode()` function for implementation details.
 */
export function TransformEncode(schema, references, value) {
    return Visit(schema, references, value);
}
